import java.util.List;
import java.io.File;
import java.io.IOException;

import org.jdom2.Document;
import org.jdom2.Element;
import org.jdom2.Namespace;
import org.jdom2.JDOMException;


/**
 * Class that allows for manipulating and outputting a .urdf.xacro file based on
 * featuremodel info.
 */
public class UrdfAdapter extends Adapter {
	//Variables used exclusively by .urdf.xacro adapter
	public final String in_uxName;
	public final Document urdfModel;
	public final Namespace NS;
	
	/**
	 * Creates a runnable Adapter for .urdf.xacro files.
	 * 
	 * When reading input in_uxPath may contain other files that will be filtered out.
	 * 
	 * @throws IOException 
	 * @throws JDOMException 
	 */
	public UrdfAdapter() throws JDOMException, IOException {
		super();
		
		//Get the selected .urdf.xacro file name as String
		in_uxName = botName+UX_ENDING;
		
		//Get the selected .urdf.xacro file as Document
		List<String> fileMatchList = XMLHelper.findFile(in_uxName, in_uxPath);
		
		if(fileMatchList.isEmpty()) {
			System.err.println("Cannot handle \'"+in_uxName+"\': No input file known.");
			System.exit(-1);
		}
		if(fileMatchList.size()!=1) {
			System.err.println("Cannot handle \'"+in_uxName+"\': Multiple input files found:");
			for(String match : fileMatchList) {
				System.err.println("\'"+match+"\'");
			}
			System.exit(-1);
		}
		
		urdfModel = XMLHelper.readXml(fileMatchList.get(0));
		
		//Get the "xacro" namespace of the .urdf.xacro document as Namespace
		NS = urdfModel.getRootElement().getNamespace("xacro");
	}

	/**
	 * Run the Adapter for .urdf.xacro files.
	 * 
	 * Sets the {@code <xacro:include>} tags.
	 * The links/ joints are not edited, as the precise visual bot representation is not
	 * important for this project.
	 * 
	 * @throws IOException 
	 */
	public void run() throws IOException {
		System.out.println("UrdfAdapter started.");
		
		//Set <xacro:include> tags
		this.setInclude();
		
		switch(this.mode) {
			case "NO_CHANGE":
				break;
			case "Custom":
				/**
				 * If the urdf.xacro would be modified (to have an accurate visual model with the sensors attached to it), 
				 * this needs to happen here. (Delete or Insert <link> and <joint> tags.) 
				 * This would require separate files with <link> and <joint> tags for each bot and sensor for the Adapter to access.
				 *  
				 * For example, like this:
				 * The "autogenerated"/"other" directory gets one child directory for each bot. 
				 * Each of these new directories contains one file for each sensorconfig with the building blocks 
				 * to be inserted in .urdf.xacro, just like the sensorconfigs contain building blocks 
				 * for the .gazebo.xacro files.
				 */
				break;
			default:
				System.err.println("Invalid input in \'"+this.in_fmName+"\': NO_CHANGE/ Custom parameter could not be parsed.");
				System.exit(-1);
		}
		
		//Output urdfModel
		this.output();
		
		System.out.println("UrdfAdapter finished.");
	}
	
	/**
	 * Sets {@code <xacro:include>} tags in urdfModel to correctly point to the files to be included.
	 */
	private void setInclude() {
		List<Element> includeList = XMLHelper.xQueryElements(this.urdfModel, "//xacro:include", this.NS);
		
		if(includeList.size()!=2) {
			System.err.println("Unexpected input in \'"+this.in_uxName+"\': Expected exactly 2 <xacro:include> tags.");
			System.exit(-1);
		}
		
		int gxIndex;
		int commonPropIndex;
		if(includeList.get(0).getAttributeValue("filename").endsWith(GX_ENDING)) {
			gxIndex = 0;
			commonPropIndex = 1;
		}
		else {
			gxIndex = 1;
			commonPropIndex = 0;
		}
		
		String gxDestDirectoryName = (new File(this.out_gxPath)).getName();
		includeList.get(gxIndex).setAttribute("filename", "$(find "+gxDestDirectoryName+")/"+this.out_gxName);
		
		//The common_properties.xacro stays in its original directory, so this is not edited
		//String commonPropDestDirectoryName = ;
		//includeList.get(commonPropIndex).setAttribute("filename", "$(find "+commonPropDestDirectoryName+")/common_properties.xacro");
	}
	
	/**
	 * Output the urdfModel as XML file.
	 * 
	 * @throws IOException 
	 */
	private void output() throws IOException {
		XMLHelper.saveAsXml(this.urdfModel, this.out_uxPath+this.out_uxName);
	}
}
